#include "itemset.h"
extern int myId, pNum;

ITLSTNODE * newItsLnkLstNode( int itNum, int occCount )
{
  ITLSTNODE * lst;

  lst=( ITLSTNODE * ) malloc ( sizeof ( ITLSTNODE ) );
  lst->next=NULL;
  lst->idSet=NULL;
  lst->itemCount=itNum;
  lst->occurCount=occCount;

  return lst;
}

/* allocate space for a new item set */
void newItemSet( ITEMSET * newSet, int setSize)
{
  newSet->size=setSize;
  newSet->idSet= ( int * ) malloc ( setSize*sizeof( int ) );
  newSet->currentIdx=0;
  newSet->support=0;
  newSet->recCount=0;
  newSet->recSet=NULL;
}

void copyItemSet( ITEMSET * dst, ITEMSET * src )
{
  dst->size=src->size;
  dst->idSet=src->idSet;
  dst->currentIdx=src->currentIdx;
  dst->support=src->support;
  dst->recCount=src->recCount;
  dst->recSet=src->recSet;
}

/* add an item into an item set */
void addAnItem( ITEMSET * set, int id )
{
  if( set->currentIdx>=set->size ) {
    printf("number of items greater than expected \n ");
    exit( 1 );
  }
  set->idSet[ ( set->currentIdx )++ ]=id;
}

/*
 * lexically compare two itemset's ids 
 */
int compItemSets( ITEMSET * is0, ITEMSET * is1 )
{
  /* if ( is0->size != is1->size ) return 0; */
  /* if ( is0->currentIdx != is1->currentIdx ) return 0; */
  /* i=(is0->currentIdx); */
  return ( memcmp( (is0->idSet), (is1->idSet), (is0->size)*sizeof( int ) ) );
}

/* lexically compare two id sets */
int compIdLst( int * idl0, int * idl1, int count )
{
  return ( memcmp( idl0,  idl1, count*sizeof( int )));
}

/* when two item sets match, it means:
 * 1) the item number they contain are same;
 * 2) each item in the set are identical except the last one
 * if two itemset match, then we combine them to a new itemset
 * which contain these two itemsets' joint set
 */
int combineItemSet( int * table, int colNum, ITEMSET * is0, ITEMSET * is1 , ITEMSET * is2 )
{
  int i, j, newRecNum, coOccurs, newId;
  int * newIdSet;
  int * newRecSet;

  if ( ( is0->size ) != ( is1->size ) ) {
    return 0;
  }

  i=(is0->size);

  if( ( is0->idSet )[i-1]==( is1->idSet )[i-1] ) {
    return 0;
  }

  newIdSet = ( int * ) malloc ( ( ( is0->size)+1 )*sizeof( int ) ) ;
  is2->idSet=newIdSet;
  is2->size=(is0->size)+1 ;

  for ( i=0; i<( is0->size)-1; i++ ) {
    if ( ( is0->idSet )[i]==( is1->idSet )[i] ) {
      newIdSet[ i ]=( is0->idSet )[ i ];
    } else {
      break;
    }
  }

  if( i<(is0->size)-1 ) {
    free( newIdSet );
    is2->currentIdx=0;
    is2->idSet=NULL;
    return 0;
  }

  /* i==( ( is0->size )-1 ) here */

  newIdSet[ i ] = ( is0->idSet )[ i ];
  newIdSet[ i+1 ] = ( is1->idSet )[ i ];
  is2->currentIdx=i+1;
  is2->support++;
  is2->size=( is0->size )+1;

  /*
   * we don't know how many records contain the itemset in both
   * itemset 0 and itemset 1, but we know it will not exceed the
   * smaller number of records in either one, so we simplly request 
   * memory enough for the smaller set
   */
  newId=is2->idSet[ i ];
  newRecNum= ( is0->recCount<is1->recCount ? is0->recCount : is1->recCount );
  newRecSet = ( int * ) malloc ( newRecNum*sizeof( int ) );

  /*
   * calculating how many times the item with id "newId"
   * occurs in the records of is0
   */
  coOccurs=occurTimes( table, colNum, is0, newId, newRecSet );
  if ( coOccurs==0 ) {
    free( newIdSet );
    is2->currentIdx=0;
    is2->idSet=NULL;
    free( newRecSet );
    is2->recCount=0;
    return 0;
  }

  is2->recCount=coOccurs;
  is2->recSet=newRecSet;

  return 1;
}

int occurTimes( int * table, int colNum, ITEMSET* is0, int id, int * recSet )
{
  int coOccurTimes=0;
  int i;

  for( i=0; i<is0->recCount; i++ ) {
    if( table[ ( is0->recSet )[i]*colNum+id ] ) {
      coOccurTimes++;
      recSet[ i ]=( is0->recSet )[i];
    }
  }

  return coOccurTimes;
}

/*
 * count how many occurence happened in database
 * in the new item set
 * since is1 contains one more item than is0 
 * and item set of is0 is subset of is1,
 * we just need to count how many records
 * containing is0 also contains the last 
 * item in is1 
 */
int occurCount( int * table, int colNum, ITEMSET * is0, ITEMSET * is1 )
{
  int tempCount;
  int i, itemId;
  int * newRecSet;
  int j;

  itemId=is0->idSet[ ( is1->size )-1 ];
  i=is0->recCount;

/* printf(" is0's item number %d\n", is0->size );
 * printf("is0: (");

 * for( j= 0; j<is0->recCount; j++ ) {
 * printf("%d,",is0->recSet[j]);
 * }
 * printf(")");
 */

newRecSet = ( int * ) malloc( ( is0->recCount )*sizeof( int ) ) ;

  is1->recSet = ( int * ) malloc( ( is0->recCount )*sizeof( int ) ) ;
  while( i-->0 ) {
    if ( table[ is0->recSet[i]*colNum+itemId ]>0 ) {
      ( is1->recCount )++;
      is1->recSet[i]=is0->recSet[i];
    }
  }

 return is1->recCount;
}


/* delete an item set */
void delItemSet( ITEMSET * set )
{
  free( set->idSet );
  set->idSet=NULL;
  set->size=0;
  set->currentIdx=0;
  set->support=0;
  set->recCount=0;
  free (set->recSet);
  set->recSet=NULL;
}

/*
 * allocate space for an item set list 
 * contain the same number of items
 */
void newItemSetList( ITEMSETLIST * itStLst, int itemNum, int size )
{
  ITEMSET * ip;

  itStLst->itemNum=itemNum;
  itStLst->listSize=size;
  itStLst->currentIdx=0;

  ip=( ITEMSET * ) malloc ( size*sizeof( ITEMSET ) );
  itStLst->setList=ip;
}

/* add an itemset in an item set list */
void addAnItemSet( ITEMSETLIST * list )
{
  if ( list->currentIdx>=list->listSize ) {
    printf("number of itemsets in a set list is greater than expected \n ");
    exit( 1 );
  }

  newItemSet( ( list->setList )+( list->currentIdx), list->itemNum );
  ( list->currentIdx )++;
}

/* remove an item set list */
void delItemSetList( ITEMSETLIST * list )
{
  ITEMSET * sp;
   
  sp=( list->setList )+( list->currentIdx );
  while( sp--!=list->setList ) {
    delItemSet( sp );
  }
  free( list->setList );
  list->setList=NULL;
}

/*
 * this list is used to contain the real min-itemset
 * list which means every item set in it is with the
 * support greater than the minimum requirement
 */
MINITEMS * newMinItemSetList( int size )
{
  MINITEMS * newMinSetList;
  int i;

  newMinSetList = ( MINITEMS * ) malloc ( sizeof ( MINITEMS ) );

  /*
   * I suppose for each i, there will be at least one itemset of
   * the size i , i ranges from 1 to the number of the items
   * so newMinSetList[i] contains the pointer to the list of all
   * item set with size i.
   */

  newMinSetList->listNum=size;
  newMinSetList->itemSetListOfSize
    = ( ITEMSETLIST * )malloc( size*sizeof ( ITEMSETLIST ) );
  newMinSetList->currentIdx=0;
  return newMinSetList;
}

void addItemSetList( MINITEMS * minSetList )
{
  int itNum, thisSize, nextSize;

  if ( minSetList->currentIdx>=minSetList->listNum ) {
    printf("number of itemset list is greater than expected \n ");
    exit( 1 );
  }

  itNum=( ( ( minSetList->itemSetListOfSize)[minSetList->currentIdx] ).itemNum )+1;
  /*
   * assuming that if the size of previous item set is n
   * then the size of this item set is n*(n-1)/2
   * maximum expectation
   */
  thisSize=( minSetList[ minSetList->currentIdx ] ).currentIdx;
  nextSize=thisSize*(thisSize-1)/2;

  newItemSetList( &( 
    ( minSetList->itemSetListOfSize )[minSetList->currentIdx] ), itNum, nextSize );
  minSetList->currentIdx++;
}

/*
 * remove the min-itemset list
 */
void delMinItemSetList( MINITEMS * setList )
{
  ITEMSETLIST * setListIdx;

  setListIdx=( setList->itemSetListOfSize )
    +( setList->currentIdx );
  while( setListIdx-- != setList->itemSetListOfSize ) {
    delItemSetList( setListIdx );
  }
  free ( setList->itemSetListOfSize );
  setList->itemSetListOfSize = NULL;
}

/* MINITEMS * newMinItSetList( DBASE * db ) */
ITLSTNODE * newMinItSetList( int * table, int rowNum, int colNum )
{
  MINITEMS * minSetp;
  int *sumOfFields, *tempBuf, *bufCurs;
  int i,j,k,tempCount;
  ITEMSET *cp;
  ITLSTNODE * minLnkLstHead;

  minLnkLstHead=newItsLnkLstNode( 0,0 );
  minSetp=newMinItemSetList( colNum );

  sumOfFields=(int *) malloc ( colNum*sizeof( int ) );

  /* initialize the number of items in each itemset list
   * for i>0, minSetp[i] means the list of item set whose
   * number of items in the set is `i'
   * minSet[0] is the list of `original' item set, which means
   * the set of all items in database before generating 
   * the `real' item set with all sizes
   */
  for( i=0; i<colNum; i++ ) {
    ( ( minSetp->itemSetListOfSize)[i] ).itemNum=i;
  }

  ( ( minSetp->itemSetListOfSize )[0] ).listSize=colNum;
  /* newItemSetList( minSetp->itemSetListOfSize, 1, db->numOfItems); */

  initMining( table, rowNum, colNum, minSetp );

  rootGetItemSets( minLnkLstHead, minSetp, 1 );
  MPI_Barrier( MPI_COMM_WORLD );

  if ( myId==0 ) {
    tempCount=sortMinLst( minLnkLstHead, 1 );
  }

  for ( i=2; i<=colNum; i++ ) {
    updateRel( table, colNum, minSetp, i );
    rootGetItemSets( minLnkLstHead, minSetp, i );

    if ( myId==0 ) {
      /* how many item sets occur more than minimum requirement */
      tempCount=sortMinLst( minLnkLstHead, i );
    }

    MPI_Bcast( &tempCount, 1 , MPI_INT, 0, MPI_COMM_WORLD );
    tempBuf=(int *) malloc( tempCount*i*sizeof( int ) );

    if( myId==0 ) {
      copyIds( minLnkLstHead,tempBuf,i );
    }

/* printf("id%d: break 4, i=%d\n",myId, i); */

    MPI_Bcast( tempBuf, i*tempCount , MPI_INT, 0, MPI_COMM_WORLD );

    /* for ( j=0; j<tempCount; j++ ) {
     *  MPI_Bcast( tempBuf, i , MPI_INT, 0, MPI_COMM_WORLD );
     * } */

    /* MPI_Barrier( MPI_COMM_WORLD ); */
    itemSetLocalUpdt( tempBuf, minSetp, i, tempCount );

  }

  /* return minSetp; */
  return( minLnkLstHead );
}

void itemSetLocalUpdt( int * isets, MINITEMS * minSetp, int setSize, int setCount )
{
  ITEMSETLIST * list0;
  ITEMSET * is0;
  int i, j, moveAhead;
  int *setCursor0;

  list0=( ( minSetp->itemSetListOfSize )+( minSetp->currentIdx )-1 );
  is0=list0->setList;

  for( i=0; i<list0->currentIdx; i++,is0++ ) {

    for( j=0, setCursor0=isets; j<setCount; j++ ) {
      if ( compIdLst( is0->idSet, setCursor0, setSize )==0 ) {
        break;
      } else {
        setCursor0 += setSize;
        continue;
      }
    }

    /*
     * if the item set in local node doesn't appear
     * in the item list distributed by root node
     * remove it
     */
    if( j>=setCount ) {
      is0->recCount=0;
    }
  }

  moveAhead=0;

  for( i=0, is0=list0->setList; i<list0->currentIdx; i++,is0++ ) {
    if( moveAhead>0 ) {
      copyItemSet( is0-moveAhead, is0 );
    }

    if( is0->recCount==0 ) {
      moveAhead++;
    }
  }

  ( list0->currentIdx ) -= moveAhead;
}

rootGetItemSets( ITLSTNODE * headNode, MINITEMS * minSetp, int numOfItems )
{
  int i, j, k, position=0 ;
  ITEMSETLIST *curISetListp, *prevItemSetList ;
  ITEMSET *curItemp;
  char buffer[ LOCALBUFSIZE ], rootBuf[ LOCALBUFSIZE ] ;
  int *ids, subCount, *lSize;
  MPI_Request *sReq, *rReq;
  MPI_Status status;

/* printf("id%d: break 1: numOfItems=%d, (minSetp->currentIdx)=%d \n"
 * ,myId, numOfItems, (minSetp->currentIdx) );
 */

  curISetListp = ( minSetp->itemSetListOfSize )+( (minSetp->currentIdx)-1 );
  curItemp = (curISetListp->setList);
  lSize=( int * ) malloc( pNum*sizeof( int ) );

  MPI_Gather( &(curISetListp->currentIdx), 1, MPI_INT
    , lSize, 1, MPI_INT, 0, MPI_COMM_WORLD );

  /* send current item set list to root */
  for( i=0; i<( curISetListp->currentIdx ); i++ ) {

  /* prepare to send the size of itemset, itemset's ids, 
   * and the occurence to root
   */
    position=0;
    MPI_Pack( curItemp->idSet, curItemp->size, MPI_INT
      , buffer, LOCALBUFSIZE, & position, MPI_COMM_WORLD );
    MPI_Pack( & (curItemp->recCount), 1, MPI_INT
      , buffer, LOCALBUFSIZE, & position, MPI_COMM_WORLD );
    MPI_Send( buffer, position, MPI_PACKED
      , 0, 0, MPI_COMM_WORLD );
    curItemp++;
    /* format of the data to be sent:
     * at the beginning is a list of ids, then, the 
     * number of records in which the ids occur together.
     */ 
  }

  if ( myId==0 ) {

    for( j=0; j<pNum; j++ ) {
      /* MPI_Recv( &lSize, 1, MPI_INT, j, 0
       * , MPI_COMM_WORLD, &status );
       */
      for( k=0; k<lSize[ j ]; k++ ) {
        MPI_Recv( rootBuf, LOCALBUFSIZE, MPI_PACKED, j
          , 0, MPI_COMM_WORLD, &status );
        position=0;
        ids=(int *)malloc( numOfItems*sizeof( int ) );
        MPI_Unpack( rootBuf, LOCALBUFSIZE, &position,
          ids, numOfItems, MPI_INT, MPI_COMM_WORLD );
        MPI_Unpack( rootBuf, LOCALBUFSIZE, &position,
          &subCount, 1 , MPI_INT, MPI_COMM_WORLD );
        registerItemSet( headNode, numOfItems, ids, subCount );
      }
    }
  }

  free( lSize );
  lSize=NULL;
}

/*
 * if the id set can be found, add up the record count
 * if not create a new one and attach to the list
 */
registerItemSet( ITLSTNODE * itLnkLst, int numOfItems, int * ids, int subCount )
{
  int i,j,k, compResult;
  ITLSTNODE *node0, *newNode;

  node0=itLnkLst;

  while( node0->next ) {
    /* find the beginning of the item set list with the size of "numOfItems" */
    if( ( ( node0->next )->itemCount )<numOfItems ) {
      node0=node0->next;
      continue;
    }

    /* if no item list of size "numOfItems", exit */

    if( ( ( node0->next )->itemCount )>numOfItems ) {
      break;
    }

    /* if the item list of size "numOfItems" found, add up the record
     * count
     */
    compResult=compIdLst( ( node0->next )->idSet, ids, numOfItems ); 
    if( compResult==0 ) {
      ( ( node0->next )->occurCount)+=subCount;
      node0=node0->next;
      return;
    }
    /* continue finding */
    if( compResult<0 ) {
      node0=node0->next;
      continue;
    } else { /* if( compResult>0 ) */
      /* no such item list exist, exit */
      break;
    }
  }

  /* insert a node */
  newNode=newItsLnkLstNode( numOfItems, subCount );
  newNode->idSet=ids;
  newNode->next=node0->next;
  node0->next=newNode;
}

/* if the occurence of an itemset is below the minimum requirement
 * delete it. And keep the itemsets whose occurence is above the 
 * minimum support. return value is the count of such itemsets
 */
int sortMinLst( ITLSTNODE * headNode, int itemCount )
{
  ITLSTNODE *temp, *temp1;
  int i=0;

  temp=headNode;

  while ( temp->next ) {
    if ( ( ( temp->next )->itemCount )<itemCount ) {
      temp=temp->next;
      continue;
    }

    if ( ( ( temp->next )->itemCount )>itemCount )
      break;

    if (( ( temp->next )->itemCount )==itemCount ) {

      if ( ((temp->next)->occurCount)<MINISUPPORT ) {
        temp1=temp->next;
        temp->next=temp1->next;
        /* temp=temp->next; */
        free( temp1 );
      } else {
        i++;
        temp=temp->next;
      }
      continue;
    }
  }

/* if(itemCount==2)
 * printf("int sortMinLst i=%d\n", i);
 */

  return i;
}

void copyIds( ITLSTNODE * headNode, int * buf , int itemNum )
{
  ITLSTNODE *temp;
  int *temp1;

  temp=headNode;
  temp1=buf;

  while( temp->next && ( temp->next )->itemCount<itemNum) {
    temp=temp->next;
  }

  while( temp->next ) {
    /* temp1=( int * )strncpy(( char* )temp1, (char *)((temp->next)->idSet), itemNum*sizeof( int ) ); */
    temp1=memcpy(temp1, ((temp->next)->idSet), itemNum*sizeof( int ) );
    temp=temp->next;
    temp1=temp1+itemNum;
  }
}

/*
 * In this function, we initialize the base of item set 
 * which means to put each item in a seperate itemset
 */
void initMining( int * table, int recNum, int itemNum, MINITEMS * minSetp )
{
  int i=0,j=0,k=0, tempSum=0;
  int * recLst;
  ITEMSETLIST * curList;
  ITEMSET * curSet;

  newItemSetList( minSetp->itemSetListOfSize, 1, itemNum );
  curList=minSetp->itemSetListOfSize;

  for( i=0; i<itemNum; i++ ) {
    tempSum=0;
    k=0;
    recLst=( int * ) malloc( recNum*sizeof( int ) );
    for( j=0; j<recNum; j++ ) {
      if( table[ j*itemNum+i ]>0 ) {
        recLst[ k++ ]=j;
        tempSum+=table[ j*itemNum+i ];
      }
    }

    if( tempSum>1 ) {
      curSet=( curList->setList )+(curList->currentIdx);
      newItemSet( curSet , 1 );
      ( curSet->idSet )[ 0 ]=i;
      curSet->support = tempSum;
      curSet->recSet=recLst;
      curSet->recCount=k;
      (curList->currentIdx)++;
    } else {
      free( recLst );
      recLst=NULL;
    }
  }
  minSetp->currentIdx=1;
}

ITEMSETLIST * updateRel( int * table, int colNum, MINITEMS * minSetp, int t )
{
  int i, j, k, prevSize;
  ITEMSETLIST * curISetListp, * prevItemSetList ;

  curISetListp = ( minSetp->itemSetListOfSize )+( minSetp->currentIdx );
  prevItemSetList=curISetListp-1;
  prevSize=prevItemSetList->currentIdx;
  newItemSetList( curISetListp, t, prevSize*(prevSize-1)/2 );

  newItemSet( curISetListp->setList, t );

  for ( i=0; i<prevSize; i++ ) {

    if( ( ( prevItemSetList->setList )[ i ] ).recCount==0 ) {
      continue;
    }

    for ( j=i+1; j<prevSize; j++ ) {

      if( ( ( prevItemSetList->setList )[ j ]).recCount==0 ) {
        continue;
      }

      if( combineItemSet( table, colNum,
        &( ( prevItemSetList->setList )[ i ] ),
        &( ( prevItemSetList->setList )[ j ] ),
        &( ( curISetListp->setList )[ curISetListp->currentIdx ] ) ) ) {

          if( ( (curISetListp->setList[ curISetListp->currentIdx ]).recCount )
            >1 ) {
          /* if(((curISetListp->setList[ curISetListp->currentIdx ]).recCount )
           * >(double)MINISUPPORT/(double)pNum )
           */
            ( curISetListp->currentIdx )++;
          } else {
            delItemSet( (curISetListp->setList)+( ( curISetListp->setList )->currentIdx) );
          }
      }
    }
  }
  ( minSetp->currentIdx )++;
}

void showRelat( MINITEMS * minList )
{
  int i,j,k ;
  ITEMSETLIST * curItmLst;
  ITEMSET * curItm;

  for( i=0; i<minList->currentIdx; i++ ) {
    printf( "item set with %d elements \n", i+1 );
    curItmLst=( minList->itemSetListOfSize )+i;
    for( j=0; j<curItmLst->currentIdx; j++ ) {
      curItm=(curItmLst->setList)+j;
      printf("(");
      for( k=0; k<curItm->size; k++ ) {
        printf( "%d,",curItm->idSet[k] );
      }
      printf(")\t");
    }
    printf("\n");
  }
}

void showMinItSets( ITLSTNODE * minSetLnkLst )
{
  ITLSTNODE * curs0;
  int setSize=0, i;

  curs0=minSetLnkLst->next;
  if ( curs0 ) {
    printf("The min-item sets found: \n");
  } else {
    printf("No min-item set found \n");
    return;
  }

  while( curs0 ) {
    if( setSize!=curs0->itemCount ) {
      setSize=curs0->itemCount;
      printf("\n min-itemset with size %d:\n", setSize);
    } 

    printf("(");
    for ( i=0; i<setSize; i++ )
      printf("%d,",( curs0->idSet )[i]);
    printf(")\n");

    curs0=curs0->next;
  }
}
